/// <reference types="node" />
/// <reference types="node" />
import net from "net";
import tls from "tls";
import { NetConn } from "./netConn";
import { Logger } from "./logger";
/**
 * NetService is a TCP or TLS server that listen on a port and create a NetConn for each connection
 */
export declare class NetService {
    private TAG;
    private logger?;
    private server;
    private serverType;
    private port;
    private tlsOptions?;
    private options?;
    private connClass;
    private serviceName;
    static DEBUG: boolean;
    private waitPromise?;
    /**
     * Create a new NetService
     * @param port port to listen
     * @param connClass A class that extends NetConn
     * @param tlsOptions option for tls.createServer leave undefined for tcp server
     * @param options optional options for the NetConn
     * @param logger optional logger
     */
    constructor(port: number, connClass: NetConnClass, tlsOptions?: tls.TlsOptions, options?: any, logger?: Logger);
    /**
     * Print a log message
     * @param msg
     */
    log(msg: string, err?: Error): void;
    /**
     * Listen for connections
     * @returns A promise that resolves when the server is listening
     */
    listen(): Promise<void>;
    /**
     * Close the server and stop listening for connections
     */
    close(): void;
    /**
     * Handle a new connection. Create a new NetConn and add event handlers
     * @param socket
     */
    private onConnection;
    /**
     * Handle server close event
     */
    onClose(): void;
    /**
     * Handle server error event. If there is a waitPromise reject it
     * @param error
     */
    onError(error: any): void;
}
export type NetConnClass = {
    new (socket: net.Socket, server?: any, options?: any, logger?: Logger): NetConn;
};
export type WaitPromise = {
    resolve: () => void;
    reject: (error: Error) => void;
};
